#!/usr/bin/env python

# very basic robot controller 
# requires to be running: amcl, gazebo robot, rviz):  roslaunch sim amcl

import os.path
import rospy
import sys, traceback, getopt
from geometry_msgs.msg import Twist
from geometry_msgs.msg import PoseArray
from nav_msgs.msg import OccupancyGrid
from sensor_msgs.msg import LaserScan
from nav_msgs.msg import Odometry, Path
from std_srvs.srv import Empty
from tf import TransformListener
from geometry_msgs.msg import Pose, PoseStamped, PointStamped, PoseWithCovarianceStamped, Point, Quaternion, Twist
import tf
import math
import geometry_msgs.msg
import random
import pickle


startGlobalLocalize = False
linSpeed = 0.5        # linear speed
rotSpeed = 0.5         # totation speed
minObstacleDist = 0.7  # minimum obstacle distance
minParticlesDist = 8   # minimum particles distance

# frames
TOPIC_MAP        = 'map'                    # fixed frame
TOPIC_ROBOT      = 'base_link'              # robot base frame
TOPIC_PARTICLES  = 'particlecloud'          # amcl particles topic
TOPIC_PERIM      = 'perim'                  # rviz perimeter path
# messages
TOPIC_ODOM       = 'odom'                   # odometry topic
TOPIC_NAVI       = 'cmd_vel_mux/input/navi' # robot's motor velocity topic
TOPIC_GLOCALIZE  = 'global_localization'    # amcl trigger global localization
TOPIC_SCAN       = 'laserscan'              # laserscan topic for obstacle detection
TOPIC_RVIZ_CLICK = 'clicked_point'          # rviz publish point topic

PERIM_FILE = "perim.pkl"

MODE_PERIM = 0
MODE_MOW   = 1

STATE_OFF             = 0
STATE_GLOBAL_LOCALIZE = 1
STATE_MOW             = 2


# scale to -PI..PI
def scalePI(v):
  d = v
  while (d < 0):
    d=d+2*math.pi
  while (d >= 2*math.pi):
    d=d-2*math.pi
  if (d >= math.pi):
    return (-2*math.pi+d)
  elif (d < -math.pi):
    return (2*math.pi+d)
  else:
    return d

# distance between current (x) and set-value (w)
def distancePI(x, w):
  d = scalePI(w-x)
  if (d < -math.pi):
    d=d+2*math.pi
  elif (d>math.pi):
    d=d-2*math.pi
  return d


def rand(vmin, vmax):
  v = vmin + random.random() * (vmax - vmin)
  return v


class Control():
  def __init__(self):
        # initiliaze
        if startGlobalLocalize:
          self.state = STATE_GLOBAL_LOCALIZE
        else: 
          self.state = STATE_MOW
        self.r_particles = 9999
        self.pose = None
        self.counter = 0
        self.rmin = 0
        self.rmax = 0
        self.rfront = 0
        self.perim = Path()
        self.seq = 1
        self.mapReceived = False
        self.odoReceived = False
        self.pose = Pose()

        rospy.init_node('Control', anonymous=False)

	# tell user how to stop TurtleBot
	rospy.loginfo("To stop TurtleBot CTRL + C")

        # What function to call when you ctrl + c    
        rospy.on_shutdown(self.shutdown)

        self.cmd_perim = rospy.Publisher(TOPIC_PERIM, Path, queue_size=1)
        self.cmd_vel = rospy.Publisher(TOPIC_NAVI, Twist, queue_size=10)
        self.cmd_global_localization = rospy.ServiceProxy(TOPIC_GLOCALIZE, Empty)
        self.__scan_sub = rospy.Subscriber(TOPIC_SCAN, LaserScan, self.__scan_handler)
        self.__map_sub = rospy.Subscriber(TOPIC_MAP, OccupancyGrid, self.__map_handler)
        self.__click_sub = rospy.Subscriber(TOPIC_RVIZ_CLICK, PointStamped, self.__click_handler)
        self.__particle_sub = rospy.Subscriber(TOPIC_PARTICLES, PoseArray, self.__particle_handler)                    
        
	# Create a publisher which can "talk" to TurtleBot and tell it to move
        # Tip: You may need to change cmd_vel_mux/input/navi to /cmd_vel if you're not using TurtleBot2
        self.tf = TransformListener()  
        rospy.sleep(1)  
        t = self.tf.getLatestCommonTime(TOPIC_ROBOT, TOPIC_MAP)        
        self.tf.waitForTransform(TOPIC_MAP, TOPIC_ROBOT, t, rospy.Duration(1.0))      

        self.__scan_odom_sub = rospy.Subscriber(TOPIC_ODOM, Odometry, self.__odom_handler)
        while not self.odoReceived:
          rospy.sleep(0.5)
        print "start"

	#TurtleBot will stop if we don't keep telling it to move.  How often should we tell it to move? 10 HZ
        #r = rospy.Rate(5);
        
        # Twist is a datatype for velocity
        if self.state == STATE_GLOBAL_LOCALIZE:
          self.cmd_global_localization()
          twist = Twist()
          twist.linear.x = 0
          twist.angular.z = 0.5
          self.cmd_vel.publish(twist)            

        self.timeout = rospy.Time.now()
        self.nextPathTime = rospy.Time.now()
        #rospy.sleep(2)  

        self.load()        
        
	# as long as you haven't ctrl + c keeping doing...
        while not rospy.is_shutdown():	                
          if self.state == STATE_GLOBAL_LOCALIZE:              
            # rotate robot until particle radius is small
            if self.r_particles < minParticlesDist:                
              self.state = STATE_MOW
              twist = Twist()
              twist.linear.x = 0
              twist.angular.z = 0
              self.cmd_vel.publish(twist)            
          elif mode == MODE_MOW:
            if self.state == STATE_MOW:
              self.mow()
          rospy.sleep(0.1)


  def mow(self):
       if self.counter == 0: 
         return
       now = rospy.Time.now()
       #type(pose) = geometry_msgs.msg.Pose              
       print "rfront=", self.rfront, "p=", self.pose.position.x, self.pose.position.y, "y=", self.yaw
       twist = Twist()
       twist.linear.x = linSpeed
       twist.angular.z = 0
       self.cmd_vel.publish(twist)            
       if now < self.timeout:                
         return
       # act on laser range
       if self.rfront < minObstacleDist:                
         print "obstacle"         
         self.turnDegree(rand(math.pi-math.pi/4, math.pi+math.pi/4))
         #self.linearDuration(linSpeed, 1)
         now = rospy.Time.now()
         #self.timeout = now + rospy.Duration(1)                  
         # act on position
       if (not self.isInside(self.pose.position)):
         print "end-of-world"                  
         self.turnUntilInside()
         self.linearDuration(linSpeed, 1)         
         now = rospy.Time.now()
         #self.timeout = now + rospy.Duration(1)              

  def turnDegree(self, angle):
      dyaw = scalePI(self.yaw + angle) 
      print "turnDegree ", self.yaw/math.pi*180, dyaw/math.pi*180
      twist = Twist()
      while (True):
        twist.linear.x = 0;        
        self.cmd_vel.publish(twist)            
        delta = distancePI(self.yaw, dyaw)/math.pi*180
        if (delta > 0): 
          twist.angular.z = rotSpeed
        else:
          twist.angular.z = -rotSpeed
        print "delta=", delta, " rfront=", self.rfront
        if (abs(delta) < 15):
          break
        rospy.sleep(0.2)
      twist.angular.z = 0
      self.cmd_vel.publish(twist)            
      rospy.sleep(1)
   
  def turnUntilInside(self):
      print "turnUntilInside"
      twist = Twist()
      if rand(-1, 1) > 0:
        twist.angular.z = rotSpeed
      else:
        twist.angular.z = -rotSpeed
      while True:
        pt = self.pose.position
        pt.x = self.pose.position.x+math.cos(self.yaw)*0.5
        pt.y = self.pose.position.y+math.sin(self.yaw)*0.5
        if self.isInside(pt):
          break
        twist.linear.x = 0.0;
        self.cmd_vel.publish(twist)            
        rospy.sleep(0.2)   
      stopTime = rospy.get_time() + 2;
      while (rospy.get_time() < stopTime):
        self.cmd_vel.publish(twist)            
        rospy.sleep(0.2)   
      twist.angular.z = 0
      self.cmd_vel.publish(twist)            
      rospy.sleep(1)


  def turnDuration(self, duration):
      print "turnDuration"
      twist = Twist()
      currentTime = rospy.get_time();
      stopTime = rospy.get_time() + duration;
      while (rospy.get_time() < stopTime):
         twist.linear.x = 0.0;
         twist.angular.z = rotSpeed
         self.cmd_vel.publish(twist)            
         rospy.sleep(0.2)
      twist.angular.z = 0
      self.cmd_vel.publish(twist)            
      rospy.sleep(1)


  def linearDuration(self, x, duration):
      print "linearDuration"
      twist = Twist()
      currentTime = rospy.get_time();
      stopTime = rospy.get_time() + duration;
      while (rospy.get_time() < stopTime):
         twist.linear.x = x; 
         twist.angular.z = 0
         self.cmd_vel.publish(twist)            
         rospy.sleep(0.2)
  
  def __map_handler(self, msg):
        self.mapReceived = True
        print "map received"


  def __scan_handler(self, msg):
        global counter
        self.current_laser_msg = msg        
        msg = self.current_laser_msg
        rngs = msg.ranges
        self.rfront = rngs[200]
        self.rmin = min(rngs)
        self.rmax = max(rngs)
	self.counter = self.counter + 1
        #print "rmin=", self.rmin

  def onUpdateRobotPose(self):
        quaternion = (self.pose.orientation.x,self.pose.orientation.y,self.pose.orientation.z,self.pose.orientation.w)
        euler = tf.transformations.euler_from_quaternion(quaternion)
        self.roll = euler[0]
        self.pitch = euler[1]
        self.yaw = euler[2]       
        now = rospy.Time.now()
 

  def addPerim(self, pt):
        print "addPerim", pt.x, pt.y        
        pose = PoseStamped()
        pose.header.frame_id = TOPIC_MAP
        pose.pose.position.x = pt.x
        pose.pose.position.y = pt.y
        pose.pose.position.z = float(self.pose.position.z)
        pose.pose.orientation.x = float(self.pose.orientation.x)
        pose.pose.orientation.y = float(self.pose.orientation.y)
        pose.pose.orientation.z = float(self.pose.orientation.z)
        pose.pose.orientation.w = float(self.pose.orientation.w)
        pose.header.seq = self.seq
        self.perim.header.frame_id = TOPIC_MAP
        self.perim.header.stamp = rospy.Time.now()
        pose.header.stamp = self.perim.header.stamp
        self.perim.poses.append(pose)
        #print self.perim
        self.seq = self.seq + 1
        self.cmd_perim.publish(self.perim)


  def __odom_handler(self, msg):    
        #self.pose = msg.pose.pose        
        t = self.tf.getLatestCommonTime(TOPIC_ROBOT, TOPIC_MAP)        
        p, o = self.tf.lookupTransform(TOPIC_MAP, TOPIC_ROBOT, t)             
        self.pose.position.x = p[0]
        self.pose.position.y = p[1]
        self.pose.position.z = p[2]
        self.pose.orientation.x = o[0]
        self.pose.orientation.y = o[1]
        self.pose.orientation.z = o[2]
        self.pose.orientation.w = o[3]
        #print self.pose
        #print self.orientation
        self.onUpdateRobotPose()    
        self.odoReceived = True

  def __click_handler(self, pt):
        pose = Pose(pt.point,Quaternion(0.0,0.0,0.0,1.0))
        #print "click ", pt.point
        self.addPerim(pt.point)
        self.save()

  def __particle_handler(self, msg):
        if self.state != STATE_GLOBAL_LOCALIZE:
          return
        xmin = 99999
        xmax = 0
        ymin = 99999
        ymax = 0
        for pose in msg.poses:
          xmin = min(xmin, pose.position.x)
          xmax = max(xmax, pose.position.x)
          ymin = min(ymin, pose.position.y)
          ymax = max(ymax, pose.position.y)
        xr = abs(xmax-xmin)
        yr = abs(ymax-ymin)
        self.r_particles = max(xr, yr)
        print "particles ", xr, yr 

  def isInside(self, pt):
    c = False
    npts = len(self.perim.poses)
    j = npts-1
    for i in range(0, npts):
      #print i, j
      if (  ((self.perim.poses[i].pose.position.y>pt.y) != (self.perim.poses[j].pose.position.y>pt.y)) and (pt.x < (self.perim.poses[j].pose.position.x-self.perim.poses[i].pose.position.x) * (pt.y-self.perim.poses[i].pose.position.y)/(self.perim.poses[j].pose.position.y-self.perim.poses[i].pose.position.y)+self.perim.poses[i].pose.position.x)  ):
        c = not c
      j=i
    return c


  def say(self, msg):
        """Prints a message to the screen!"""
        print(msg)
        sys.stdout.flush()

  def save(self):
    print "save ", PERIM_FILE, " pts ", len(self.perim.poses)
    with open(PERIM_FILE, 'wb') as f:
       pickle.dump(self.perim, f)
        
  def load(self):
    if os.path.exists(PERIM_FILE):      
      print "load ", PERIM_FILE
      with open(PERIM_FILE, 'rb') as f:
       self.perim = pickle.load(f)
    print "pts ", len(self.perim.poses)
    self.cmd_perim.publish(self.perim)
        
  def shutdown(self):
        # stop turtlebot
        rospy.loginfo("Stop TurtleBot")
	# a default Twist has linear.x of 0 and angular.z of 0.  So it'll stop TurtleBot
        self.cmd_vel.publish(Twist())
	# sleep just makes sure TurtleBot receives the stop command prior to shutting down the script
        rospy.sleep(1)
 
if __name__ == '__main__':
   try:
     while True:
       print "NOTE: add perimeter points via rviz PublishPoint tool, enable path '/perim' to see perimeter"
       mode = input("choose run mode (0=add perimeter points, 1=mow):  ")
       print mode
       if mode >= 0 and mode <= 1:
         break
     Control()
   except:
     rospy.loginfo("Control node terminated.")
     traceback.print_exc()

